@@ -2,6 +2,10 @@
 
 /* Copyright 1993-1999 Werner Almesberger. See file COPYING for details. */
 
+#ifndef _GNU_SOURCE
+#define _GNU_SOURCE
+#endif
+
 #ifdef HAVE_CONFIG_H
 #include <config.h>
 #endif
@@ -23,8 +27,13 @@
 #include <sys/stat.h>
 #include <netinet/in.h>
 #include <arpa/inet.h>
+#ifdef __linux__
 #include <linux/kdev_t.h>	/* for MKDEV */
 #include <linux/major.h>	/* for LOOP_MAJOR */
+#endif
+#include <libintl.h>
+#include <locale.h>
+#define _(String) gettext (String)
 
 #include "comm.h"
 #include "loop.h"         /* for loop_info */
@@ -42,6 +51,14 @@
 #define LOOP_MAJOR 7
 #endif
 
+#ifndef MAJOR
+#define MAJOR(arg) 6           /* something that doesn't = LOOP_MAJOR */
+#endif
+
+#ifndef MKDEV
+#define MKDEV(arg1, arg2) mknod("/dev/Isuck", arg1, arg2) /* this is wrong */
+#endif
+
 
 #define REF_FILE   1		/* an open file */
 #define REF_ROOT   2		/* current root */
@@ -58,9 +75,9 @@
 
 typedef struct _net_cache
 {
+  struct sockaddr_storage rmt_addr;
   int lcl_port;
   int rmt_port;
-  unsigned long rmt_addr;
   ino_t ino;
   struct _net_cache *next;
 }
@@ -140,11 +157,14 @@ static int ipv4only = 0, ipv6only = 0;
 
 
 static void
-parse_net_file (SPACE_DSC * dsc,char *filename, NET_CACHE **lastptr )
+parse_net_file (SPACE_DSC * dsc,char *filename, NET_CACHE **lastptr,int version )
 {
   FILE *file;
   NET_CACHE *new, *last;
   char line[MAX_LINE + 1];
+  char rmt_addr[128];
+  char addr6[128];
+
   if (!(file = fopen (filename, "r")))
     {
       perror (filename);
@@ -160,13 +180,26 @@ parse_net_file (SPACE_DSC * dsc,char *filename, NET_CACHE **lastptr )
 	  perror ("malloc");
 	  exit (1);
 	}
-      if (sscanf (line, "%*d: %*x:%x %lx:%x %*x %*x:%*x %*x:%*x %*x %*d %*d "
-		  "%ld", &new->lcl_port, &new->rmt_addr, &new->rmt_port,
+      if (sscanf (line, "%*d: %*x:%x %64[0-9A-Fa-f]:%x %*x %*x:%*x %*x:%*x %*x %*d %*d "
+		  "%Ld", &new->lcl_port, rmt_addr, &new->rmt_port,
 		  &new->ino) != 4)
 	{
 	  free (new);
 	  continue;
 	}
+      if (strlen(rmt_addr) > 8) {
+        sscanf(rmt_addr, "%08X%08X%08X%08X",
+            &((struct sockaddr_in6 *)&new->rmt_addr)->sin6_addr.s6_addr32[0],
+            &((struct sockaddr_in6 *)&new->rmt_addr)->sin6_addr.s6_addr32[1],
+            &((struct sockaddr_in6 *)&new->rmt_addr)->sin6_addr.s6_addr32[2],
+            &((struct sockaddr_in6 *)&new->rmt_addr)->sin6_addr.s6_addr32[3]);
+        inet_ntop(AF_INET6, &((struct sockaddr_in6 *)&new->rmt_addr)->sin6_addr, addr6, sizeof(addr6));
+        printf("address %s\n", addr6);
+      } else {
+        sscanf(rmt_addr, "%X",
+            &((struct sockaddr_in *) &new->rmt_addr)->sin_addr.s_addr);
+            ((struct sockaddr *) &new->rmt_addr)->sa_family = AF_INET;
+      }
       if (!new->ino)
 	{
 	  free (new);
@@ -187,9 +220,11 @@ parse_net_file (SPACE_DSC * dsc,char *filename, NET_CACHE **lastptr )
 static void
 fill_net_cache (SPACE_DSC * dsc)
 {
-  char buffer[PATH_MAX + 1];
   NET_CACHE *last;
   struct stat statbuf;
+  char *buffer = malloc (strlen (PROC_BASE) + strlen (dsc->name) + 8);
+  if (!buffer)
+    return;
 
   if (dsc->once)
     return;
@@ -205,18 +240,18 @@ fill_net_cache (SPACE_DSC * dsc)
     sprintf (buffer, PROC_BASE "/net/%s", dsc->name);
     if (stat(buffer, &statbuf) != 0) {
       if (ipv4only) 
-        fprintf(stderr, "-4 flag used but proc file %s is not readable/\n", buffer);
+        fprintf(stderr, _("-4 flag used but proc file %s is not readable\n"), buffer);
     } else {
-      parse_net_file (dsc, buffer, &last );
+      parse_net_file (dsc, buffer, &last,4 );
     }
   }
   if (!ipv4only) {
     sprintf (buffer, PROC_BASE "/net/%s6", dsc->name);
     if (stat(buffer, &statbuf) != 0) {
       if (ipv6only) 
-        fprintf(stderr, "-6 flag used but proc file %s is not readable/\n", buffer);
+        fprintf(stderr, _("-6 flag used but proc file %s is not readable\n"), buffer);
     } else {
-      parse_net_file (dsc, buffer, &last );
+      parse_net_file (dsc, buffer, &last,6 );
     }
   }
 }
@@ -229,7 +264,7 @@ fill_unix_cache (void)
   FILE *file;
   UNIX_CACHE *new, *last;
   struct stat st;
-  char path[PATH_MAX + 1], line[MAX_LINE + 1];
+  char *path = NULL, line[MAX_LINE + 1];
   int ino;
 
   if (once)
@@ -244,10 +279,14 @@ fill_unix_cache (void)
   (void) fgets (line, MAX_LINE, file);
   while (fgets (line, MAX_LINE, file))
     {
-      if (sscanf (line, "%*x: %*x %*x %*x %*x %*x %d %s", &ino, path) != 2)
+      if (sscanf (line, "%*x: %*x %*x %*x %*x %*x %d %as", &ino, &path) != 2)
 	continue;
-      if (stat (path, &st) < 0)
+      if (stat (path, &st) < 0) {
+	free (path);
 	continue;
+      }
+      free (path);
+
       new = malloc (sizeof (UNIX_CACHE));
       new->fs_dev = st.st_dev;
       new->fs_ino = st.st_ino;
@@ -398,15 +437,16 @@ check_dir (const char *rel, pid_t pid, int type)
 {
   DIR *dir;
   struct dirent *de;
-  char path[PATH_MAX + 1];
+  char *path;
 
   if (!(dir = opendir (rel)))
     return;
   while ((de = readdir (dir)) != NULL)
     if (strcmp (de->d_name, ".") && strcmp (de->d_name, ".."))
       {
-	sprintf (path, "%s/%s", rel, de->d_name);
+	asprintf (&path, "%s/%s", rel, de->d_name);
 	check_link (path, pid, type);
+	free (path);
       }
   (void) closedir (dir);
 }
@@ -417,7 +457,7 @@ scan_fd (void)
 {
   DIR *dir;
   struct dirent *de;
-  char path[PATH_MAX + 1];
+  char *path;
   pid_t pid;
   int empty;
 
@@ -431,7 +471,8 @@ scan_fd (void)
     if ((pid = atoi (de->d_name)) != 0)
       {
 	empty = 0;
-	sprintf (path, "%s/%d", PROC_BASE, pid);
+	if (asprintf (&path, "%s/%d", PROC_BASE, pid) < 0)
+	  continue;
 	if (chdir (path) >= 0)
 	  {
 	    check_link ("root", pid, REF_ROOT);
@@ -442,11 +483,12 @@ scan_fd (void)
 	    check_map ("maps", pid, REF_MMAP);
 	    check_dir ("fd", pid, REF_FILE);
 	  }
+	free (path);
       }
   (void) closedir (dir);
   if (empty)
     {
-      fprintf (stderr, PROC_BASE " is empty (not mounted ?)\n");
+      fprintf (stderr, _("%s is empty (not mounted ?)\n"),PROC_BASE);
       exit (1);
     }
 }
@@ -457,18 +499,21 @@ scan_mounts (void)
 {
   FILE *file;
   struct stat st_dev, st_parent, st_mounted;
-  char line[MAX_LINE + 1], path[PATH_MAX + 1], mounted[PATH_MAX + 3];
+  char line[MAX_LINE + 1], *path = NULL, *mounted = NULL;
   char *end;
 
   if (!(file = fopen (PROC_BASE "/mounts", "r")))
     return;			/* old kernel */
   while (fgets (line, MAX_LINE, file))
     {
-      if (sscanf (line, "%s %s", path, mounted) != 2)
+      if (sscanf (line, "%as %as", &path, &mounted) != 2)
 	continue;
       /* new kernel :-) */
-      if (stat (path, &st_dev) < 0)
-	continue;		/* might be NFS or such */
+      if (stat (path, &st_dev) < 0) {
+	free (path);            /* might be NFS or such */
+	free (mounted);
+	continue;		
+      }
       if (S_ISBLK (st_dev.st_mode) && MAJOR (st_dev.st_rdev) == LOOP_MAJOR)
 	{
           struct loop_info loopinfo;
@@ -484,16 +529,25 @@ scan_mounts (void)
       if (stat (mounted, &st_mounted) < 0)
 	{
 	  perror (mounted);
+	  free (path);
+	  free (mounted);
 	  continue;
 	}
-      end = strchr (mounted, 0);
-      strcpy (end, "/..");
-      if (stat (mounted, &st_parent) >= 0)
+      if (asprintf (&end, "%s/..", mounted) < 0)
+	{
+	  free (path);
+	  free (mounted);
+	  continue;
+	}
+
+      if (stat (end, &st_parent) >= 0)
 	{
-	  *end = 0;
 	  add_other (it_mount, st_parent.st_dev, st_mounted.st_dev,
 		     st_mounted.st_ino, mounted);
 	}
+      free (end);
+      free (path);
+      free (mounted);
     }
   (void) fclose (file);
 }
@@ -504,19 +558,22 @@ scan_swaps (void)
 {
   FILE *file;
   struct stat st;
-  char line[MAX_LINE + 1], path[PATH_MAX + 1], type[MAX_LINE + 1];
+  char line[MAX_LINE + 1], *path, type[MAX_LINE + 1];
 
   if (!(file = fopen (PROC_BASE "/swaps", "r")))
     return;			/* old kernel */
   (void) fgets (line, MAX_LINE, file);
   while (fgets (line, MAX_LINE, file))
     {
-      if (sscanf (line, "%s %s", path, type) != 2)
+      if (sscanf (line, "%as %s", &path, type) != 2)
 	continue;		/* new kernel :-) */
-      if (strcmp (type, "file"))
+      if (strcmp (type, "file")) {
+	free (path);
 	continue;
+      }
       if (stat (path, &st) >= 0)
 	add_other (it_swap, st.st_dev, st.st_dev, st.st_ino, path);
+      free (path);
     }
   (void) fclose (file);
 }
@@ -528,20 +585,19 @@ ask (pid_t pid)
   int ch, c;
 
   fflush (stdout);
-  do
-    {
-      fprintf (stderr, "Kill process %d ? (y/n) ", pid);
-      fflush (stderr);
-      do
-	if ((ch = getchar ()) == EOF)
-	  exit (0);
-      while (ch == '\n' || ch == '\t' || ch == ' ');
-      do
-	if ((c = getchar ()) == EOF)
-	  exit (0);
-      while (c != '\n');
-    }
-  while (ch != 'y' && ch != 'n' && ch != 'Y' && ch != 'N');
+  do {
+    fprintf (stderr, _("Kill process %d ? (y/n) "), pid);
+    fflush (stderr);
+    do {
+      if ((ch = getchar ()) == EOF)
+        exit (0);
+      if (ch == '\n') return 0;
+    } while (ch == '\t' || ch == ' ');
+    do
+      if ((c = getchar ()) == EOF)
+	    exit (0);
+    while (c != '\n');
+  } while (ch != '\n' && ch != 'y' && ch != 'n' && ch != 'Y' && ch != 'N');
   return ch == 'y' || ch == 'Y';
 }
 
@@ -560,19 +616,19 @@ kill_item (const FILE_DSC * file, const ITEM_DSC * item)
 	return;
       if (kill (item->u.proc.pid, file->sig_num) >= 0)
 	break;
-      sprintf (tmp, "kill %d", item->u.proc.pid);
+      sprintf (tmp, _("kill %d"), item->u.proc.pid);
       perror (tmp);
       break;
     case it_mount:
-      fprintf (stderr, "No automatic removal. Please use  umount %s\n",
+      fprintf (stderr, _("No automatic removal. Please use  umount %s\n"),
 	       item->u.misc.path);
       break;
     case it_loop:
-      fprintf (stderr, "No automatic removal. Please use  umount %s\n",
+      fprintf (stderr, _("No automatic removal. Please use  umount %s\n"),
 	       item->u.misc.path);
       break;
     case it_swap:
-      fprintf (stderr, "No automatic removal. Please use  swapoff %s\n",
+      fprintf (stderr, _("No automatic removal. Please use  swapoff %s\n"),
 	       file->name);
       break;
     }
@@ -587,7 +643,7 @@ show_files_or_kill (void)
   FILE *f;
   const struct passwd *pw;
   const char *user, *scan;
-  char tmp[10], path[PATH_MAX + 1], comm[COMM_LEN + 1];
+  char tmp[10], *path, comm[COMM_LEN + 1];
   int length, header, first, dummy;
   header = 1;
   for (file = files; file; file = file->next)
@@ -657,13 +713,14 @@ show_files_or_kill (void)
 		switch (item->type)
 		  {
 		  case it_proc:
-		    sprintf (path, PROC_BASE "/%d/stat", item->u.proc.pid);
+		    asprintf (&path, PROC_BASE "/%d/stat", item->u.proc.pid);
 		    strcpy (comm, "???");
 		    if ((f = fopen (path, "r")) != NULL)
 		      {
 			(void) fscanf (f, "%d (%[^)]", &dummy, comm);
 			(void) fclose (f);
 		      }
+		    free (path);
 		    name = comm;
 		    uid = item->u.proc.uid;
 		    break;
@@ -674,7 +731,7 @@ show_files_or_kill (void)
 		    uid = 0;
 		    break;
 		  default:
-		    fprintf (stderr, "Internal error (type %d)\n",
+		    fprintf (stderr, _("Internal error (type %d)\n"),
 			     item->type);
 		    exit (1);
 		  }
@@ -704,13 +761,13 @@ show_files_or_kill (void)
 			    !(item->u.proc.ref_set & REF_EXE) ? 'm' : '.');
 		    break;
 		  case it_mount:
-		    printf ("kernel mount  ");
+		    printf (_("kernel mount  "));
 		    break;
 		  case it_loop:
-		    printf ("kernel loop   ");
+		    printf (_("kernel loop   "));
 		    break;
 		  case it_swap:
-		    printf ("kernel swap   ");
+		    printf (_("kernel swap   "));
 		    break;
 		  }
 		if (name)
@@ -730,8 +787,7 @@ show_files_or_kill (void)
 	if (!(file->flags & FLAG_VERB) || first)
 	  putchar ('\n');
 	if (first)
-	  fprintf (stderr, "No process references; use -v for the complete"
-		   " list\n");
+	  fprintf (stderr, _("No process references; use -v for the complete list\n"));
 	if (file->flags & FLAG_KILL)
 	  for (item = file->items; item; item = item->next)
 	    kill_item (file, item);
@@ -792,10 +848,10 @@ enter_item (const char *name, int flags, int sig_number, dev_t dev,
 
 static int
 parse_inet (const char *spec, const char *name_space, int *lcl_port,
-	    unsigned long *rmt_addr, int *rmt_port)
+	    struct sockaddr_storage *rmt_addr, int *rmt_port)
 {
   char *s, *here, *next, *end;
-  int port, field;
+  int port, field, address_match;
 
   if (!(s = strdup (spec)))
     {
@@ -803,8 +859,9 @@ parse_inet (const char *spec, const char *name_space, int *lcl_port,
       exit (1);
     }
   *lcl_port = *rmt_port = -1;
-  *rmt_addr = 0;
+  memset(rmt_addr, 0, sizeof(struct sockaddr_storage));
   field = 0;
+  address_match = 0;
   for (here = s; here; here = next ? next + 1 : NULL)
     {
       next = strchr (here, ',');
@@ -834,16 +891,20 @@ parse_inet (const char *spec, const char *name_space, int *lcl_port,
 	case 1:
 	  if (!*here)
 	    break;
-	  if ((long) (*rmt_addr = inet_addr (here)) == -1)
-	    {
-	      struct hostent *hostent;
+          if (!ipv4only) {
 
-	      if (!(hostent = gethostbyname (here)))
-		return 0;
-	      if (hostent->h_addrtype != AF_INET)
-		return 0;
-	      memcpy (rmt_addr, hostent->h_addr, hostent->h_length);
-	    }
+            if (inet_pton(AF_INET6, here, &((struct sockaddr_in6*)rmt_addr)->sin6_addr) > 0) {
+              address_match = 1;
+              rmt_addr->ss_family = AF_INET6;
+             }
+          }
+          if (!ipv6only && !address_match) {
+            if (inet_pton(AF_INET, here, &((struct sockaddr_in*)rmt_addr)->sin_addr) > 0) {
+              address_match = 1;
+              rmt_addr->ss_family = AF_INET6;
+            }
+          }
+            
 	  break;
 	default:
 	  return 0;
@@ -864,7 +925,7 @@ static void find_net_dev(void)
   }
   if (fd >= 0)
     close(fd);
-  fprintf(stderr,"can't find sockets' device number");
+  fprintf(stderr,_("can't find sockets' device number"));
 }
 
 
@@ -872,39 +933,38 @@ static void find_net_dev(void)
 static void
 usage (void)
 {
-  fprintf (stderr, "usage: fuser [ -a | -s ] [ -n space ] [ -signal ] "
-	   "[ -kimuv ] name ...\n%13s[ - ] [ -n space ] [ -signal ] [ -kimuv ] "
-	   "name ...\n", "");
-  fprintf (stderr, "       fuser -l\n");
-  fprintf (stderr, "       fuser -V\n\n");
-  fprintf (stderr, "    -a        display unused files too\n");
-  fprintf (stderr, "    -k        kill processes accessing that file\n");
-  fprintf (stderr, "    -i        ask before killing (ignored without -k)\n");
-  fprintf (stderr, "    -l        list signal names\n");
-  fprintf (stderr, "    -m        mounted FS\n");
-  fprintf (stderr, "    -n space  search in the specified name space (file, "
-	   "udp, or tcp)\n");
-  fprintf (stderr, "    -s        silent operation\n");
-  fprintf (stderr, "    -signal   send signal instead of SIGKILL\n");
-  fprintf (stderr, "    -u        display user ids\n");
-  fprintf (stderr, "    -v        verbose output\n");
-  fprintf (stderr, "    -V        display version information\n");
-  fprintf (stderr, "    -4        search IPv4 sockets only\n");
-  fprintf (stderr, "    -6        search IPv6 sockets only\n");
-  fprintf (stderr, "    -         reset options\n\n");
-  fprintf (stderr, "  udp/tcp names: [local_port][,[rmt_host][,[rmt_port]]]"
-	   "\n\n");
+  fprintf (stderr, _(
+    "usage: fuser [ -a | -s ] [ -n space ] [ -signal ] [ -kimuv ] name ...\n"
+    "             [ - ] [ -n space ] [ -signal ] [ -kimuv ] name ...\n"
+    "       fuser -l\n"
+    "       fuser -V\n\n"
+    "    -a        display unused files too\n"
+    "    -k        kill processes accessing that file\n"
+    "    -i        ask before killing (ignored without -k)\n"
+    "    -l        list signal names\n"
+    "    -m        mounted FS\n"
+    "    -n space  search in the specified name space (file, udp, or tcp)\n"
+    "    -s        silent operation\n"
+    "    -signal   send signal instead of SIGKILL\n"
+    "    -u        display user ids\n"
+    "    -v        verbose output\n"
+    "    -V        display version information\n"
+    "    -4        search IPv4 sockets only\n"
+    "    -6        search IPv6 sockets only\n"
+    "    -         reset options\n\n"
+    "  udp/tcp names: [local_port][,[rmt_host][,[rmt_port]]]\n\n"));
   exit (1);
 }
 
 void print_version()
 {
-  fprintf(stderr, "fuser (psmisc) %s\n", VERSION);
-  fprintf(stderr, "Copyright (C) 1993-2000 Werner Almesberger and Craig Small\n\n");
-  fprintf(stderr, "PSmisc comes with ABSOLUTELY NO WARRANTY.\n");
-  fprintf(stderr, "This is free software, and you are welcome to redistribute it under the terms\n");
-  fprintf(stderr, "of the GNU General Public License.\n");
-  fprintf(stderr, "For more information about these matters, see the files named COPYING.\n");
+  fprintf(stderr, _("fuser (psmisc) %s\n"), VERSION);
+  fprintf(stderr, _(
+    "Copyright (C) 1993-2002 Werner Almesberger and Craig Small\n\n"
+    "PSmisc comes with ABSOLUTELY NO WARRANTY.\n"
+    "This is free software, and you are welcome to redistribute it under the terms\n"
+    "of the GNU General Public License.\n"
+    "For more information about these matters, see the files named COPYING.\n"));
 }
 
 
@@ -913,13 +973,18 @@ int
 main (int argc, char **argv)
 {
   SPACE_DSC *name_space;
-  char path[PATH_MAX + 1];
   int flags, silent, do_kill, sig_number, no_files;
 
   flags = silent = do_kill = 0;
   sig_number = SIGKILL;
   name_space = name_spaces;
   no_files = 1;
+
+  /* Setup the i18n */
+  setlocale(LC_ALL, "");
+  bindtextdomain(PACKAGE, LOCALEDIR);
+  textdomain(PACKAGE);
+
   if (argc < 2)
     usage ();
   if (argc == 2 && !strcmp (argv[1], "-l"))
@@ -1042,7 +1107,6 @@ main (int argc, char **argv)
 		  st.st_dev = st.st_rdev;
 		else if (S_ISDIR (st.st_mode))
 		  {
-		    sprintf (path, "%s/.", *argv);
 		    if (stat (*argv, &st) < 0)
 		      {
 			perror (*argv);
@@ -1069,12 +1133,12 @@ main (int argc, char **argv)
 	  else
 	    {
 	      NET_CACHE *walk;
-	      unsigned long rmt_addr;
+	      struct sockaddr_storage rmt_addr;
 	      int lcl_port, rmt_port;
 
 	      if (flags & FLAG_DEV)
 		{
-		  fprintf (stderr, "ignoring -m in name space \"%s\"\n",
+		  fprintf (stderr, _("ignoring -m in name space \"%s\"\n"),
 			   this_name_space->name);
 		  flags &= ~FLAG_DEV;
 		}
@@ -1082,13 +1146,16 @@ main (int argc, char **argv)
 	      if (!parse_inet (*argv, this_name_space->name, &lcl_port,
 			       &rmt_addr, &rmt_port))
 		{
-		  fprintf (stderr, "%s/%s: invalid specificiation\n", *argv,
+		  fprintf (stderr, _("%s/%s: invalid specificiation\n"), *argv,
 			   this_name_space->name);
 		  continue;
 		}
 	      for (walk = this_name_space->cache; walk; walk = walk->next)
 		if ((lcl_port == -1 || walk->lcl_port == lcl_port) &&
-		    (!rmt_addr || walk->rmt_addr == rmt_addr) &&
+		    (rmt_addr.ss_family = 0|| ( memcmp(
+                     &((struct sockaddr_in6*)&walk->rmt_addr)->sin6_addr,
+                     &((struct sockaddr_in6*)&rmt_addr)->sin6_addr,
+                     sizeof(struct in6_addr)) == 0) ) &&
 		    (rmt_port == -1 || walk->rmt_port == rmt_port))
 		  enter_item (*argv, flags, sig_number, net_dev, walk->ino,
 			      this_name_space);
