@@ -17,24 +17,29 @@
 #include <dirent.h>
 #include <termios.h>
 #include <termcap.h>
+#include <langinfo.h>
+#include <locale.h>
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <sys/ioctl.h>
 
 #include "comm.h"
 
+#ifdef FLASK_LINUX
+#include <fs_secure.h>
+#endif /*FLASK_LINUX*/
 
 #ifndef MAX_DEPTH
 #define MAX_DEPTH    100
 #endif
 #define PROC_BASE    "/proc"
 
-/* UTF-8 defines by Johan Myreen */
-#define UTF_V	"\342\224\202\277"	/* Vertical line drawing char */
-#define UTF_VR	"\342\224\234\277"	/* Vertical and right */
-#define UTF_H	"\342\224\200\277"	/* Horizontal */
-#define UTF_UR	"\342\224\224\277"	/* Up and right */
-#define UTF_HD	"\342\224\254\277"	/* Horizontal and down */
+/* UTF-8 defines by Johan Myreen, updated by Ben Winslow */
+#define UTF_V	"\342\224\202"	/* U+2502, Vertical line drawing char */
+#define UTF_VR	"\342\224\234"	/* U+251C, Vertical and right */
+#define UTF_H	"\342\224\200"	/* U+2500, Horizontal */
+#define UTF_UR	"\342\224\224"	/* U+2514, Up and right */
+#define UTF_HD	"\342\224\254"	/* U+252C, Horizontal and down */
 
 #define VT_BEG	"\033(0\017"	/* use graphic chars */
 #define VT_END	"\033(B"	/* back to normal char set */
@@ -51,6 +56,9 @@ typedef struct _proc
   int argc;			/* with -a   : number of arguments, -1 if swapped    */
   pid_t pid;
   uid_t uid;
+#ifdef FLASK_LINUX
+  security_id_t sid;
+#endif /*FLASK_LINUX*/
   int highlight;
   struct _child *children;
   struct _proc *parent;
@@ -98,6 +106,10 @@ static PROC *list = NULL;
 static int width[MAX_DEPTH], more[MAX_DEPTH];
 static int print_args = 0, compact = 1, user_change = 0, pids = 0, by_pid = 0,
   trunc = 1;
+#ifdef FLASK_LINUX
+static int show_sids    = 0;
+static int show_scontext = 0;
+#endif /*FLASK_LINUX*/
 static int output_width = 132;
 static int cur_x = 1;
 static char last_char = 0;
@@ -147,6 +159,38 @@ out_int (int x)			/* non-negative integers only */
   return digits;
 }
 
+#ifdef FLASK_LINUX
+static void 
+out_sid ( security_id_t sid )
+{
+  if ( (int) sid >= 0 )
+    out_int((int) sid);
+  else
+    out_string("??");
+}
+
+static void 
+out_scontext ( security_id_t sid )
+{
+  static char buf[256];
+  int security_sid_to_context();
+  int len = sizeof(buf);
+  int rv;
+
+  bzero(buf,256);
+
+  rv = security_sid_to_context((int)sid, buf, &len);
+  if ( rv ) {
+    out_string("`??'"); /* punt */
+  }
+  else {
+    out_string("`");
+    out_string(buf);
+    out_string("'");
+  }
+}
+#endif /*FLASK_LINUX*/
+
 
 static void
 out_newline (void)
@@ -170,9 +214,13 @@ find_proc (pid_t pid)
   return walk;
 }
 
-
+#ifdef FLASK_LINUX
+static PROC *
+new_proc(const char *comm, pid_t pid, uid_t uid, security_id_t sid)
+#else  /*FLASK_LINUX*/
 static PROC *
 new_proc (const char *comm, pid_t pid, uid_t uid)
+#endif /*FLASK_LINUX*/
 {
   PROC *new;
 
@@ -185,6 +233,9 @@ new_proc (const char *comm, pid_t pid, uid_t uid)
   new->pid = pid;
   new->uid = uid;
   new->highlight = 0;
+#ifdef FLASK_LINUX
+  new->sid = sid;
+#endif /*FLASK_LINUX*/
   new->children = NULL;
   new->parent = NULL;
   new->next = list;
@@ -253,15 +304,24 @@ set_args (PROC * this, const char *args, int size)
     this->argv[i] = start = strchr (start, 0) + 1;
 }
 
-
+#ifdef FLASK_LINUX
+static void
+add_proc(const char *comm, pid_t pid, pid_t ppid, uid_t uid,
+         const char *args, int size, security_id_t sid)
+#else  /*FLASK_LINUX*/
 static void
 add_proc (const char *comm, pid_t pid, pid_t ppid, uid_t uid,
 	  const char *args, int size)
+#endif /*FLASK_LINUX*/
 {
   PROC *this, *parent;
 
   if (!(this = find_proc (pid)))
+#ifdef FLASK_LINUX
+    this = new_proc(comm, pid, uid, sid);
+#else  /*FLASK_LINUX*/
     this = new_proc (comm, pid, uid);
+#endif /*FLASK_LINUX*/
   else
     {
       strcpy (this->comm, comm);
@@ -272,7 +332,11 @@ add_proc (const char *comm, pid_t pid, pid_t ppid, uid_t uid,
   if (pid == ppid)
     ppid = 0;
   if (!(parent = find_proc (ppid)))
+#ifdef FLASK_LINUX
+    parent = new_proc("?", ppid, 0, sid);
+#else  /*FLASK_LINUX*/
     parent = new_proc ("?", ppid, 0);
+#endif /*FLASK_LINUX*/
   add_child (parent, this);
   this->parent = parent;
 }
@@ -329,8 +393,9 @@ dump_tree (PROC * current, int level, int rep, int leaf, int last,
     }
   if (current->highlight && (tmp = tgetstr ("md", NULL)))
     tputs (tmp, 1, putchar);
-  if ((swapped = print_args) && current->argc < 0)
-    out_char ('(');
+  swapped = info = print_args;
+  if (swapped  && current->argc < 0)
+    out_char ('(');  
   comm_len = 0;
   for (here = current->comm; *here; here++)
     if (*here == '\\')
@@ -350,25 +415,38 @@ dump_tree (PROC * current, int level, int rep, int leaf, int last,
 	comm_len += 4;
       }
   offset = cur_x;
-  info = pids || (user_change && prev_uid != current->uid);
-  if (info)
-    out_char (swapped ? ',' : '(');
   if (pids)
+  {
+    out_char(info++ ? ',' : '(');
     (void) out_int (current->pid);
+  }
   if (user_change && prev_uid != current->uid)
     {
-      if (pids)
-	out_char (',');
+      out_char (info++ ? ',' : '(');
       if ((pw = getpwuid (current->uid)))
 	out_string (pw->pw_name);
       else
 	(void) out_int (current->uid);
     }
-  if (info || swapped)
+#ifdef FLASK_LINUX
+  if ( show_sids ) {
+    out_char (info++ ? ',' : '(');
+    out_sid(current->sid);
+  }
+  if ( show_scontext ) {
+    out_char (info++ ? ',' : '(');
+    out_scontext(current->sid);
+  }
+#endif /*FLASK_LINUX*/
+  if ((swapped && print_args && current->argc < 0) || (!swapped && info))
     out_char (')');
   if (current->highlight && (tmp = tgetstr ("me", NULL)))
     tputs (tmp, 1, putchar);
+#ifdef FLASK_LINUX
+  if (show_scontext || print_args)
+#else  /*FLASK_LINUX*/
   if (print_args)
+#endif /*FLASK_LINUX*/
     {
       for (i = 0; i < current->argc; i++)
 	{
@@ -392,12 +470,20 @@ dump_tree (PROC * current, int level, int rep, int leaf, int last,
 	    }
 	}
     }
+#ifdef FLASK_LINUX
+  if ( show_scontext || print_args || ! current->children )
+#else  /*FLASK_LINUX*/
   if (print_args || !current->children)
+#endif /*FLASK_LINUX*/
     {
       while (closing--)
 	out_char (']');
       out_newline ();
+#ifdef FLASK_LINUX
+      if ( show_scontext || print_args )
+#else /*FLASK_LINUX*/
       if (print_args)
+#endif /*FLASK_LINUX*/
 	{
 	  more[level] = !last;
 	  width[level] = swapped + (comm_len > 1 ? 0 : -1);
@@ -480,13 +566,16 @@ read_proc (void)
   struct dirent *de;
   FILE *file;
   struct stat st;
-  char path[PATH_MAX + 1], comm[COMM_LEN + 1];
+  char *path, comm[COMM_LEN + 1];
   char *buffer;
   char readbuf[BUFSIZ+1];
   char *tmpptr;
   pid_t pid, ppid;
   int fd, size;
   int empty;
+#ifdef FLASK_LINUX
+  security_id_t sid = -1;
+#endif /*FLASK_LINUX*/
 
   if (!print_args)
     buffer = NULL;
@@ -504,12 +593,18 @@ read_proc (void)
   while ((de = readdir (dir)) != NULL)
     if ((pid = atoi (de->d_name)) != 0)
       {
+	if (!(path = malloc (strlen (PROC_BASE) + strlen (de->d_name) + 10)))
+	  exit (2);
 	sprintf (path, "%s/%d/stat", PROC_BASE, pid);
 	if ((file = fopen (path, "r")) != NULL)
 	  {
 	    empty = 0;
-            sprintf (path, "%s/%d", PROC_BASE, pid);
+	    sprintf (path, "%s/%d", PROC_BASE, pid);
+#ifdef FLASK_LINUX
+            if (fstat_secure(fileno(file),&st,&sid) < 0)
+#else /*FLASK_LINUX*/
             if (stat (path, &st) < 0)
+#endif /*FLASK_LINUX*/
 	    {
 		perror (path);
 		exit (1);
@@ -534,7 +629,11 @@ read_proc (void)
 		 &ppid) == 4)
 	      */
 		if (!print_args)
+#ifdef FLASK_LINUX
+		  add_proc(comm, pid, ppid, st.st_uid, NULL, 0, sid);
+#else  /*FLASK_LINUX*/
 		  add_proc (comm, pid, ppid, st.st_uid, NULL, 0);
+#endif /*FLASK_LINUX*/
 		else
 		  {
 		    sprintf (path, "%s/%d/cmdline", PROC_BASE, pid);
@@ -551,11 +650,18 @@ read_proc (void)
 		    (void) close (fd);
 		    if (size)
 		      buffer[size++] = 0;
+#ifdef FLASK_LINUX
+		    add_proc(comm, pid, ppid, st.st_uid, buffer, size, sid);
+#else  /*FLASK_LINUX*/
 		    add_proc (comm, pid, ppid, st.st_uid, buffer, size);
+#endif /*FLASK_LINUX*/
 		  }
-	      } } }
+		}
+	      }
+	    }
 	    (void) fclose (file);
 	  }
+	free (path);
       }
   (void) closedir (dir);
   if (print_args)
@@ -587,7 +693,11 @@ read_stdin (void)
 	cmd = comm;
       if (*cmd == '-')
 	cmd++;
+#ifdef FLASK_LINUX
+      add_proc(cmd, pid, ppid, uid, NULL, 0, -1);
+#else  /*FLASK_LINUX*/
       add_proc (cmd, pid, ppid, uid, NULL, 0);
+#endif /*FLASK_LINUX*/
     }
 }
 
@@ -612,6 +722,10 @@ usage (void)
   fprintf (stderr, "    -n     sort output by PID\n");
   fprintf (stderr, "    -p     show PIDs; implies -c\n");
   fprintf (stderr, "    -u     show uid transitions\n");
+#ifdef FLASK_LINUX
+  fprintf (stderr, "    -s     show Flask SIDs\n");
+  fprintf (stderr, "    -x     show Flask security contexts\n");
+#endif /*FLASK_LINUX*/
   fprintf (stderr,
 	   "    -U     use UTF-8 (Unicode) line drawing characters\n");
   fprintf (stderr, "    -V     display version information\n");
@@ -621,6 +735,16 @@ usage (void)
   exit (1);
 }
 
+void print_version()
+{
+  fprintf(stderr, "pstree (psmisc) %s\n", VERSION);
+  fprintf(stderr, "Copyright (C) 1993-2002 Werner Almesberger and Craig Small\n\n");
+  fprintf(stderr, "pstree comes with ABSOLUTELY NO WARRANTY.\n");
+  fprintf(stderr, "This is free software, and you are welcome to redistribute it under the terms\n");
+  fprintf(stderr, "of the GNU General Public License.\n");
+  fprintf(stderr, "For more information about these matters, see the files named COPYING.\n");
+}
+
 
 int
 main (int argc, char **argv)
@@ -638,7 +762,14 @@ main (int argc, char **argv)
   pid = 1;
   highlight = 0;
   pw = NULL;
+  
+  setlocale(LC_ALL, "");
+  
+#ifdef FLASK_LINUX
+  while ((c = getopt (argc, argv, "acGhH:npluUVsx")) != EOF)
+#else  /*FLASK_LINUX*/
   while ((c = getopt (argc, argv, "acGhH:npluUV")) != EOF)
+#endif /*FLASK_LINUX*/
     switch (c)
       {
       case 'a':
@@ -693,8 +824,16 @@ main (int argc, char **argv)
 	sym = &sym_utf;
 	break;
       case 'V':
-	fprintf (stderr, "pstree from psmisc version %s\n", VERSION);
+      print_version();
 	return 0;
+#ifdef FLASK_LINUX
+      case 's':
+        show_sids = 1;
+        break;
+      case 'x':
+        show_scontext = 1;
+        break;
+#endif /*FLASK_LINUX*/
       default:
 	usage ();
       }
@@ -712,6 +851,14 @@ main (int argc, char **argv)
   }
   if (optind != argc)
     usage ();
+  if (sym == &sym_ascii) {
+  	/*
+  	 * If the locale's charset is UTF-8, automatically
+  	 * use the UTF-8 symbols
+  	 */
+	if (!strcmp(nl_langinfo(CODESET), "UTF-8"))
+		sym = &sym_utf;
+  }
   read_proc ();
   for (current = find_proc (highlight); current; current = current->parent)
     current->highlight = 1;
